import groovy.json.JsonSlurper

import java.text.SimpleDateFormat
import java.util.stream.Collectors

//读取项目的基础配置
def jsonSlpuer = new JsonSlurper()
println "开始使用jCoder来执行项目"
def conf = jsonSlpuer.parseText(new File("jCoder.json").text)
println "配置文件：" + conf


//工作模块目录
workModuleFolder = System.getProperty("user.dir") +
        File.separator + conf.workModuleFolder +
        File.separator + "src" +
        File.separator + "main" +
        File.separator + "java";

println "workModuleFolder : ${workModuleFolder}"

//工作模块包名
workModulePackage = conf.workModulePackage
println "workModulePackage : ${workModulePackage}"

//SDK模块目录
sdkModuleFolder = System.getProperty("user.dir") +
        File.separator + conf.sdkModuleFolder +
        File.separator + "src" +
        File.separator + "main" +
        File.separator + "java";
println "sdkModuleFolder : ${sdkModuleFolder}"

//SDK模块包名
sdkModulePackage = conf.sdkModulePackage
println "sdkModulePackage : ${sdkModulePackage}"


exceptions = new ArrayList<String>();

//输入路由
println '/**'
println ' * 新版JCoder主要用于生成管理后台的crud，因此只要选择实体即可'
println ' * 我们的接口命名规划为api/[admin:管理后台接口;common:通用接口;app:WEB/APP/小程序/H5接口]'
println ' * 同理Controller的起名为[Entity]AdminController/[Entity]CommonController/[Entity]AppController'
println ' * 我们的基础思想是Request(也就是Qo)一般每个接口一个;Response一般抽象一下，分为:'
println ' * [Entity]Vo:代表通用的Vo，一般不缺少属性'
println ' * [Entity]ListVo:代表列表渲染的Vo，一般不会带json、html之类的字段'
println ' * [Entity]BaseVo:代表基础的Vo，可能比ListVo信息还要少，例如用户的BaseVo只有userId、realName、avatar信息'

println ' */'

//输入Entity
println 'Please input entity name, example : SysUserEntity'
def entityName = System.console().readLine 'Entity : '

def packageName = sdkModulePackage + ".dao.entity"
def path = new File(getPackagePath(sdkModuleFolder, packageName))
File file = new File(path, entityName + ".java");
if (!file.exists()) {
    println "ERROR. Entity ${entityName} is not found."
    return
}

entityName = entityName?.trim()

println "Your Entity is : ${entityName}, created from mysql"

def entityPath = new File(getPackagePath(sdkModuleFolder, sdkModulePackage + ".dao.entity"))
File entityFile = new File(entityPath, entityName + ".java");

generateQo(entityName, "Create");
generateQo(entityName, "Delete");
generateQo(entityName, "Edit");
generateQo(entityName, "Get");
generateQo(entityName, "Query");
generateQo(entityName, "SetStatus");
generateQo(entityName, "SetSort");

generateController(entityName);

def generateController(String entityName) {

    entityName = entityName.replaceAll("Entity", "");

    def className = entityName + "AdminController"

    def packageName = workModulePackage + ".controller.admin"

    def path = new File(getPackagePath(workModuleFolder, packageName))
    if (!path.exists()) {
        path.mkdirs();
    }

    File file = new File(path, className + ".java");
    if (file.exists())
        exceptions.add("Controller file exists. " + file.getName());
    else {
        PrintWriter output = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "utf-8"))
        output.withPrintWriter { out -> generateController(out, packageName, className, entityName) }
    }
    println "create Controller : " + file.absolutePath
    return packageName + "." + className;
}

def generateController(out, String packageName, String className, String entityName) {

    def camelEntityName = camelCase(entityName, false);
    def noPrefixEntityName = camelCase(removePrefix(camelEntityName), false);
    def UpperCaseNoPrefixEntityName =

            println "noPrefixEntityName : " + noPrefixEntityName

    out.println """package ${packageName};

import cn.hutool.core.bean.BeanUtil;
import com.qrqy.developer.common.exception.BizException;
import com.qrqy.developer.sql.BaseSqlQuery;
import com.qrqy.developer.web.result.CommonObjectResult;
import com.qrqy.developer.web.result.CommonPagingResult;
import com.qrqy.developer.web.result.ICommonResult;
import ${workModulePackage}.dao.entity.${entityName}Entity;
import ${workModulePackage}.manager.${entityName}Manager;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeanUtils;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RestController;

import javax.annotation.Resource;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @author jCoder
 * @date ${getCurrentTime()}
 */
@RestController
@Slf4j
@Validated
@Api(tags = {"xxx对象名xxx接口"}, value = "xxx对象名xxx接口")
@RequestMapping("/api/admin/${noPrefixEntityName}")
public class ${entityName}AdminController {

    @Resource
    private UserUtil userUtil;
    
    @Resource
    private ${entityName}Manager ${noPrefixEntityName}Manager;

    @ApiOperation(value = "获取xxx对象名xxx详情", response = ${entityName}Vo.class)
    @RequestMapping(value = "/get", method = RequestMethod.POST)
    public ICommonResult get(@Valid @RequestBody ${entityName}GetQo request) {

        ${entityName}Entity entity = ${noPrefixEntityName}Manager.get(request.get${firstToUpperCase(noPrefixEntityName, false)}Id());
        ${entityName}Vo vo = BeanUtil.copyProperties(entity, ${entityName}Vo.class);
        return new CommonObjectResult<>(vo);
    }

    @ApiOperation(value = "创建xxx对象名xxx", response = ${entityName}Vo.class)
    @RequestMapping(value = "/create", method = RequestMethod.POST)
    public ICommonResult create(@Valid @RequestBody ${entityName}CreateQo request) {

        ${entityName}Entity entity = BeanUtil.copyProperties(request, ${entityName}Entity.class);
        ${noPrefixEntityName}Manager.save(entity);
        ${entityName}Vo vo = BeanUtil.copyProperties(entity, ${entityName}Vo.class);

        return new CommonObjectResult<>(vo);
    }

    @ApiOperation(value = "编辑xxx对象名xxx", response = ${entityName}Vo.class)
    @RequestMapping(value = "/edit", method = RequestMethod.POST)
    public ICommonResult edit(@Valid @RequestBody ${entityName}EditQo request) {

        ${entityName}Entity entity = ${noPrefixEntityName}Manager.get(request.get${firstToUpperCase(noPrefixEntityName, false)}Id());
        if (entity == null) {
            throw BizException.DATA_NOT_FOUND;
        }
        BeanUtils.copyProperties(request, entity);
        ${noPrefixEntityName}Manager.save(entity);
        ${entityName}Vo vo = BeanUtil.copyProperties(entity, ${entityName}Vo.class);

        return new CommonObjectResult<>(vo);
    }

    @ApiOperation(value = "获取xxx对象名xxx列表（不分页）", response = ${entityName}ListVo.class)
    @RequestMapping(value = "/list", method = RequestMethod.POST)
    public ICommonResult list(@Valid @RequestBody ${entityName}QueryQo request) {

        Specification query = new BaseSqlQuery(request);
        List<${entityName}Entity> entityList = ${noPrefixEntityName}Manager.findAll(query);
        List<${entityName}ListVo> voList = BeanUtil.copyToList(entityList, ${entityName}ListVo.class);

        return new CommonObjectResult<>(voList);
    }

    @ApiOperation(value = "获取xxx对象名xxx列表（分页）", response = ${entityName}ListVo.class)
    @RequestMapping(value = "/query", method = RequestMethod.POST)
    public ICommonResult query(@Valid @RequestBody ${entityName}QueryQo request) {

        Specification query = new BaseSqlQuery(request);
        PageRequest pageable = PageRequest.of(request.getPage(), request.getSize());
        Page<${entityName}Entity> page = ${noPrefixEntityName}Manager.query(query, pageable);

        List<${entityName}ListVo> content = page.getContent().stream().map(t -> {
            ${entityName}ListVo vo = new ${entityName}ListVo();
            BeanUtils.copyProperties(t, vo);
            return vo;
        }).collect(Collectors.toList());

        return new CommonPagingResult<>(content, page);
    }


    @ApiOperation(value = "删除xxx对象名xxx", response = String.class)
    @RequestMapping(value = "/delete", method = RequestMethod.POST)
    public ICommonResult delete(@Valid @RequestBody ${entityName}DeleteQo request) {

        ${noPrefixEntityName}Manager.delete(request.get${firstToUpperCase(noPrefixEntityName, false)}Id());

        return new CommonObjectResult<>("success");
    }


    @ApiOperation(value = "设置xxx对象名xxx状态", response = String.class)
    @RequestMapping(value = "/setStatus", method = RequestMethod.POST)
    public ICommonResult setStatus(@Valid @RequestBody ${entityName}SetStatusQo request) {

        ${entityName}Entity entity = ${noPrefixEntityName}Manager.get(request.get${firstToUpperCase(noPrefixEntityName, false)}Id());
        if (entity == null) {
            throw BizException.DATA_NOT_FOUND;
        }
        entity.setStatus(request.getStatus());
        ${noPrefixEntityName}Manager.save(entity);

        return new CommonObjectResult<>("success");
    }

    @ApiOperation(value = "设置xxx对象名xxx顺序", response = String.class)
    @RequestMapping(value = "/setSort", method = RequestMethod.POST)
    public ICommonResult setSort(@Valid @RequestBody ${entityName}SetSortQo request) {

        ${entityName}Entity entity = ${noPrefixEntityName}Manager.get(request.get${firstToUpperCase(noPrefixEntityName, false)}Id());
        if (entity == null) {
            throw BizException.DATA_NOT_FOUND;
        }
        entity.setSort(request.getSort());
        ${noPrefixEntityName}Manager.save(entity);

        return new CommonObjectResult<>("success");
    }

}

"""
}


def generateQo(String entityName, String ops) {
    def qoClassName = entityName.replaceAll("Entity", ops + "Qo");
    def packageName = workModulePackage + ".model.qo"
    def path = new File(getPackagePath(workModuleFolder, packageName))
    if (!path.exists()) {
        path.mkdirs();
    }
    File file = new File(path, qoClassName + ".java");
    if (file.exists())
        exceptions.add("Qo file exists. " + file.getName());
    else {
        PrintWriter output = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "utf-8"))
        output.withPrintWriter { out -> generateQO(out, ops, packageName, qoClassName, entityName) }
    }
    println "create qo : " + file.absolutePath
    return packageName + "." + qoClassName;
}

def generateQO(out, String ops, String packageName, String qoClassName, String entityName) {
    def entityPath = new File(getPackagePath(sdkModuleFolder, sdkModulePackage + ".dao.entity"))
    File entityFile = new File(entityPath, entityName + ".java");

    def fieldList;

    out.println """package ${packageName};
import com.qrqy.developer.common.qo.IBaseQo;
import com.qrqy.developer.common.qo.PageableQo;
import io.swagger.annotations.ApiModelProperty;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.NotEmpty;
import lombok.Data;"""
    //从 Tab 复制字段
    if (ops.equals("Create")) {
        fieldList = getFields(entityFile, false)
        fieldList = fieldList.subList(2, fieldList.size() - 2)
    } else if (ops.equals("Delete")) {
        fieldList = getFields(entityFile, false)
        def subList = new ArrayList();
        subList.add(fieldList.get(0))
        subList.add("    @NotNull")
        subList.add(fieldList.get(1))
        fieldList = subList
    } else if (ops.equals("Edit")) {
        fieldList = getFields(entityFile, false)
        def subList = new ArrayList();
        subList.add(fieldList.get(0))
        subList.add("    @NotNull")
        subList.add(fieldList.get(1))
        subList.addAll(fieldList.subList(2, fieldList.size() - 2))
        fieldList = subList;

    } else if (ops.equals("Get")) {
        fieldList = getFields(entityFile, false)
        def subList = new ArrayList();
        subList.add(fieldList.get(0))
        subList.add("    @NotNull")
        subList.add(fieldList.get(1))
        fieldList = subList
    } else if (ops.equals("Query")) {
        fieldList = getFields(entityFile, false)
    } else if (ops.equals("SetStatus")) {
        fieldList = getFields(entityFile, false)
        def subList = new ArrayList();
        subList.add(fieldList.get(0))
        subList.add("    @NotNull")
        subList.add(fieldList.get(1))
        subList.add("    @ApiModelProperty(value = \"状态，1：有效；0：无效\", example = \"\")");
        subList.add("    @NotEmpty");
        subList.add("    private String status;");
        fieldList = subList
    } else if (ops.equals("SetSort")) {
        fieldList = getFields(entityFile, false)
        def subList = new ArrayList();
        subList.add(fieldList.get(0))
        subList.add("    @NotNull")
        subList.add(fieldList.get(1))
        subList.add("    @ApiModelProperty(value = \"排序，值越大越靠前\", example = \"\")");
        subList.add("    @NotNull");
        subList.add("    private Integer sort;");
        fieldList = subList
    }


    if (fieldList.stream().anyMatch(it -> it.contains("Local"))) {
        out.println "import java.time.*;"
    }
    if (fieldList.stream().anyMatch(it -> it.contains("BigDecimal"))) {
        out.println "import java.math.BigDecimal;"
    }

    out.println ""
    generateComment(out, ops, 'Qo')
    out.println "@Data"
    if (ops.equals("Query")) {
        out.println "public class $qoClassName extends PageableQo {"
    } else {
        out.println "public class $qoClassName implements IBaseQo {"
    }
    out.println ""

    //从 Tab 复制字段
    fieldList.forEach(line -> out.println line)

    out.println "}"
}


/**
 * 得到包所在的路径
 * @param workModuleFolder
 * @param packageName
 * @return
 */
def getPackagePath(String workModuleFolder, String packageName) {
    String[] vars = packageName.split("\\.")
    return workModuleFolder + File.separator + Arrays.stream(vars).collect(Collectors.joining(File.separator));
}


def getEntityIdName(entityName) {
    def entityPath = new File(getPackagePath(sdkModuleFolder, sdkModulePackage + ".dao.entity"))
    File entityFile = new File(entityPath, entityName + ".java");

    def text = entityFile.text;

    //接口路由、说明
    def matcherId = text =~ "(?ms)@Id[\\s\\S]+?private\\s+\\w+\\s+(\\w*[Ii]d)\\s*;"
    def idName = null;
    if (matcherId.find()) {
        idName = matcherId.group(1)
    }

    return idName;
}

/**
 * 从 Entity 文件获取 Field 内容，用于生成 qo 或 vo
 * @param sourceFile
 * @return
 */
def getFields(File sourceFile, boolean first) {
    def body = sourceFile.text =~ "(?ms)public\\s+class\\s+\\S+.*?\\{(.*)\\}"
    if (!body.find()) {
        println "find entity body error"
        return null;
    }
    def item = body.group(1)

    def list = new ArrayList();
    def paramMatcher = item =~ "(?ms)(/\\*\\*(.*?)\\*/)[\\s\\S]*?(private.*?;)"
    while (paramMatcher.find()) {
        def mess = paramMatcher.group(2)
        if (mess) {
            mess = mess.replaceAll("\\*", "").replaceAll("\\s", "")
            list.add("\t@ApiModelProperty(value = \"${mess}\", example = \"\")")
        }
        def field = paramMatcher.group(3)
        list.add("\t${field}\n\r")
    }
    return list.grep {
        !(it =~ /serialVersionUID/)
    }
}

def getClassShortName(classFullName) {
    String[] qoArray = classFullName.split("\\.")
    return qoArray[qoArray.length - 1]
}

def getCurrentTime() {
    new SimpleDateFormat("yyyy-MM-dd HH:mm").format(new Date())
}

def getRandom() {
    Math.abs(new Random().nextLong())
}

/**
 * 生成类注释
 * @param out
 */
def generateComment(out, ops, type) {
    out.println "/**"
    out.println " * 参数支持动态查询后缀_跟：Is、Eq、Neq、StartWith、EndWith、Like、Gt、Gte、Lt、Lte、Null、NotNull、In、NotIn"
    out.println " * @author : jCoder"
    out.println " * @date : " + new SimpleDateFormat("yyyy-MM-dd HH:mm").format(new Date())
    out.println " */"
}

/**
 * 转为驼峰命名法
 * @param name
 * @param capitalize
 * @return
 */
def camelCase(String name, boolean capitalize) {
    return name.substring(0, 1).toLowerCase() + name.substring(1);
}

/**
 * 转换为首字母大写
 * @param name
 * @param capitalize
 * @return
 */
def firstToUpperCase(String name, boolean capitalize) {
    return name.substring(0, 1).toUpperCase() + name.substring(1);
}

/**
 * 去掉前缀，一般从第二个大写字母开始
 * @param name
 */
def removePrefix(String name) {
    String sU = "ABCDEFGHIJKLMNOPQISTUVWXYZ";
    Integer pos = 0;
    for (int i = 0; i < name.length(); i++) {
        if (sU.indexOf(name.charAt(i).toString()) >= 0) {
            pos = i;
            break;
        }
    }
    String subName = name.substring(pos);
    return subName;
}